///////////////////////////////////////////////
//                                           //
// VRLab Mental Vision Module 6: camera      //
// Author: Achille Peternier                 //
// Copyright: EPFL-VRLab 2005 (C)            //
//                                           //
///////////////////////////////////////////////
#include "../data/common.h"
#pragma comment(lib, "opengl32.lib")



//////////////
// #DEFINES //
//////////////
#define MODULE_NAME "camera"



//////////
// MAIN //
//////////

int main(int argc, char *argv[])
{
   /////////////////
   // Init MVisio //
   /////////////////   
   INIT_MVISIO(MODULE_NAME, argc, argv);  
   
   // Common interface:
   USE_FONT();  
   USE_BACKGROUND(MODULE_SIZEX, MODULE_SIZEY);
   USE_QUITBUTTON(MODULE_SIZEX, MODULE_SIZEY);
   USE_CAMERA(0.0f, 4.0f, 10.0f, 0.0f, 0.0f, 0.0f);
   USE_LIGHT(0.0f, 10.0f, 0.0f);
   
   // Test:   
   MVNODE *scene = MVISIO::load("bar.MVE");   
   scene->rotate(0.0f, 90.0f, 0.0f);
   
   MVCAMERA *cameraOrto = new MVCAMERA();
   cameraOrto->setPosition(0.0f, 2.0f, 60.0f);
   cameraOrto->setTarget(0.0f, 0.0f, 0.0f);  
   
   
   ////////////////////////////
   // Point modifier interface:
   MVWINDOW *interfaceWin = new MVWINDOW();
   interfaceWin->setFont(font);
   interfaceWin->setSizes(227, 330);  
   interfaceWin->setCaption("Camera modifier");
   interfaceWin->setTexture(baseTexture);
   interfaceWin->setPosition((float) (MODULE_SIZEX/2-227/2), (float) (MODULE_SIZEY/2-330/2));		         
   
   MVBUTTON *resetButton = new MVBUTTON();
   resetButton->setFont(font);  
   resetButton->setCaption("reset");
   resetButton->setSizes(40, 20);
   resetButton->setPosition(8, 72);
   resetButton->link(interfaceWin);    
   
   // Near plane:
   MVHSCROLLBAR *modifyPointScroll1 = new MVHSCROLLBAR();   
   modifyPointScroll1->setSizeX(100);
   modifyPointScroll1->setPosition(115, 27);
   modifyPointScroll1->link(interfaceWin);
   modifyPointScroll1->setMinValue(0.0f);
   modifyPointScroll1->setMaxValue(20.0f);
   modifyPointScroll1->setValue(1.0f);   
   
   // Far plane:
   MVHSCROLLBAR *modifyPointScroll2 = new MVHSCROLLBAR();   
   modifyPointScroll2->setSizeX(100);
   modifyPointScroll2->setPosition(115, 43);
   modifyPointScroll2->link(interfaceWin);
   modifyPointScroll2->setMinValue(0.0f);   
   modifyPointScroll2->setMaxValue(100.0f);   
   modifyPointScroll2->setValue(50.0f);
   
   // FOV:
   MVHSCROLLBAR *modifyPointScroll3 = new MVHSCROLLBAR();   
   modifyPointScroll3->setSizeX(100);
   modifyPointScroll3->setPosition(115, 10);
   modifyPointScroll3->link(interfaceWin);
   modifyPointScroll3->setMinValue(0.1f);
   modifyPointScroll3->setMaxValue(179.9f);
   modifyPointScroll3->setValue(45.0f);     
   
///////////////
// Mainloop: //
///////////////

   int mouseX = 0, mouseY = 0, oldMouseX = 0, oldMouseY = 0;
   bool mouseBL = false, mouseBR = false;
   bool oldMouseBL = false;
   bool done = false;   
   float rotationX = 0.0f, rotationY = 90.0f, depth = 10.0f;
   MVID selected = 0; 
   bool occurred = false;
   MVMATERIAL *oldMaterial, *blinkMaterial;
   blinkMaterial = new MVMATERIAL();   
   blinkMaterial->setDiffuse(0.0f, 0.0f, 0.0f, 0.5f);
   blinkMaterial->setSpecular(0.0f, 0.0f, 0.0f, 0.5f);
   float blink = 0.0f;
   bool blinkDir = false;
   MVID oldSelected = 0;
   MVMATRIX camPro;
   
   MVNODE *root = new MVNODE();     
   cameraMesh->link(root);       
   light->link(cameraMesh); 
   light->setPosition(0.0f, 3.0f, 3.0f);

   float MSX = (float) MODULE_SIZEX;
   float MSY = (float) MODULE_SIZEY;
   
   MVOPENGL::cullingOff();      
      
   while (!done)
	{			      
	      // Update values:
	      root->setRotation(rotationX, rotationY, 0.0f);	  	           
	      camera->setPosition(0.0f, 5.0f, depth);	      
	      cameraMesh->setPosition(camera->getPosition());
	      
	      blinkMaterial->setAmbient(0.5f+blink, 0.05f, 0.5f, 1.0f);
	      blinkMaterial->setDiffuse(0.5f+blink, 0.5f, 0.5f, 1.0f);
	      blinkMaterial->setSpecular(0.5f+blink, 0.5f, 0.5f, 1.0f);
	      if (blinkDir)
	            blink += 1.5f * MVISIO::getFPSfactor();
	      else
	            blink -= 1.5f * MVISIO::getFPSfactor();
	      if (blink <= 0.0f)
	      {
	            blinkDir = !blinkDir;
	            blink = 0.0f;
	      }
	      else
	            if (blink >= 0.5f)
	            {
	                  blinkDir = !blinkDir;
	                  blink = 0.5f;
	            }        	      
	      
	      // Rendering background:	   
	   	MVISIO::clear(true, true, false);	   		   	
			MVISIO::begin2D();			   
			   PASS_BACKGROUND;			   
			MVISIO::end2D();					
			
			// Rendering right side:
			scene->unlink();   			
			MVOPENGL::setNearPlane(modifyPointScroll1->getValue());
			MVOPENGL::setFarPlane(modifyPointScroll2->getValue());
			MVOPENGL::setAngleOfView(modifyPointScroll3->getValue());			
         
			MVOPENGL::resize(MSX/2.0f,	MSY/17.25f, MSX/2.0f, MSY-MSY/5.55f);         
         light->setPosition(cameraMesh->getPosition());						
			MVISIO::begin3D(camera);			 
			   scene->pass();
			   //cameraMesh->pass();			   
			   light->pass();
			MVISIO::end3D();
			light->setPosition(0.0f, 3.0f, 3.0f);
			glMatrixMode(GL_PROJECTION);
			glGetFloatv(GL_PROJECTION_MATRIX, (float *)&camPro);
			glMatrixMode(GL_MODELVIEW);
			MVVECTOR3 fA = MVVECTOR3(0.0f, 0.0f, modifyPointScroll2->getValue()) * tan(E3M_TORAD(modifyPointScroll3->getValue()/2.0f));			
			MVID _selected = MVISIO::select(mouseX, mouseY); 
			if (_selected && !occurred && mouseX > MODULE_SIZEX/2)
			{
			      if (oldSelected)
			      {
			            MVNODE *_node = scene->findById(oldSelected); 
			            if (_node)
			                  ((MVMESH *)_node)->setMaterial(oldMaterial);
			      }
			      if (_selected)
			      {
			            MVNODE *_node = scene->findById(_selected);
			            if (_node)
			            {
			                  oldMaterial = ((MVMESH *)_node)->getMaterial();
			                  blinkMaterial->setTexture(0, ((MVMESH *)_node)->getMaterial()->getTexture(0));
                           ((MVMESH *)_node)->setMaterial(blinkMaterial);
                           oldSelected = _selected;
                     }
               }
			}		
			else
			      if (oldSelected)
			      {
			            MVNODE *_node = scene->findById(oldSelected);
			            if (_node)
			                  ((MVMESH *)_node)->setMaterial(oldMaterial);				
			            oldSelected = 0;
			      }
			      
			if (!occurred && mouseX > MODULE_SIZEX/2 && (mouseBL && !oldMouseBL))
			{			      
			      selected = _selected;
			      if (selected != 0)
			            camera->setTargetNode(scene->findById(selected));
			}
			
			
			// Write text:	            
         interfaceWin->clear();	            
         char text[512];
         sprintf(text, "^y0005^x0005FOV: %.1f\n^x0005Near plane: %.1f\n^x0005Far plane: %.1f\n^x0005Target: ", modifyPointScroll3->getValue(), modifyPointScroll1->getValue(), modifyPointScroll2->getValue());
         if (selected)
               strcat(text, scene->findById(selected)->getName());
         else
               strcat(text, "none");
         char matrix[256];
         MVMATRIX m = *MVCAMERA::getCameraMatrix(); //getCameraMatrix();
         sprintf(matrix, " \n\n^x0002---------------------------\n^x0005Camera matrix (OpenGL order):\n^y0005^x0020%.2f, %.2f, %.2f, %.2f\n^x0020%.2f, %.2f, %.2f, %.2f\n^x0020%.2f, %.2f, %.2f, %.2f\n^x0020%.2f, %.2f, %.2f, %.2f",
         m._11, m._12, m._13, m._14,
         m._21, m._22, m._23, m._24,
         m._31, m._32, m._33, m._34,
         m._41, m._42, m._43, m._44);         
         strcat(text, matrix);
         
         sprintf(matrix, " \n^x0002---------------------------\n^x0005Projection matrix (OpenGL order):\n^y0005^x0020%.2f, %.2f, %.2f, %.2f\n^x0020%.2f, %.2f, %.2f, %.2f\n^x0020%.2f, %.2f, %.2f, %.2f\n^x0020%.2f, %.2f, %.2f, %.2f",
         camPro._11, camPro._12, camPro._13, camPro._14,
         camPro._21, camPro._22, camPro._23, camPro._24,
         camPro._31, camPro._32, camPro._33, camPro._34,
         camPro._41, camPro._42, camPro._43, camPro._44);
         strcat(text, matrix);
         interfaceWin->write("%s", text);
         interfaceWin->format();
         
         // Compute planes vector:           	     
	      MVVECTOR3 vec_x(m.d[0], m.d[4], m.d[8]);
	      MVVECTOR3 vec_y(m.d[1], m.d[5], m.d[9]);
	      MVVECTOR3 vec_z(m.d[2], m.d[6], m.d[10]);
	
	      // Generate mouvement vector:
	      MVVECTOR3 nearP = MVVECTOR3(0.0f, 0.0f, modifyPointScroll1->getValue());
	      MVVECTOR3 farP = MVVECTOR3(0.0f, 0.0f, modifyPointScroll2->getValue());
	      MVMATRIX nearPMat = m.inverse();
			MVMATRIX farPMat = m.inverse();			
		//	nearPMat.t = nearP;
		//	farPMat.t = farP;	      
	      nearP = vec_x * nearP.x + vec_y * nearP.y + vec_z * nearP.z;
	      farP = vec_x * farP.x + vec_y * farP.y + vec_z * farP.z;	      
			
			
			// Rendering left side:			
			scene->link(root); 			
			MVOPENGL::setNearPlane(1.0f);
			MVOPENGL::setFarPlane(400.0f);
			MVOPENGL::setAngleOfView(45.0f);			
			MVOPENGL::resize(0, MSY/17.25f, MSX/2.0f, MSY-MSY/5.55f);     
			
			cameraMesh->setPosition(camera->getPosition());
			MVMATRIX cc = m.inverse();
			cc.t[0] = camera->getPosition().x;
			cc.t[1] = camera->getPosition().y;
			cc.t[2] = camera->getPosition().z;
		   cameraMesh->setPositioning(MV_MATRIX);
		   cameraMesh->setMatrix(cc);		   
		   float farZ = -modifyPointScroll2->getValue();
		   float nearZ = -modifyPointScroll1->getValue();
			cameraMesh->setRotation(camera->getRotation());
			float f = farZ * tan(E3M_TORAD(modifyPointScroll3->getValue()/2.0f));
			
			glDisable(GL_LIGHTING);	
			//MVISIO::clear(false, true, false);
			MVISIO::begin3D(cameraOrto);			   
			   glPushMatrix();
			   glDepthMask(false);			
			   glMultMatrixf((float *) &root->getFinalMatrix());  
			   glDisable(GL_CULL_FACE);
			   glEnable(GL_BLEND);
			   glColor4f(1.0f, 1.0f, 0.0f, 0.5f);
			   glMultMatrixf((float *) &farPMat);
			    
			   glBegin(GL_TRIANGLE_STRIP);
			      glVertex3f(-f, f, farZ);
			      glVertex3f(f, f, farZ);
			      glVertex3f(-f, -f, farZ);
			      glVertex3f(f, -f, farZ);
			   glEnd();
			   glPopMatrix();
			   
			   glPushMatrix();			
			   glMultMatrixf((float *) &root->getFinalMatrix());  
			   glDisable(GL_CULL_FACE);
			   glEnable(GL_BLEND);			   
			   glColor4f(1.0f, 1.0f, 0.0f, 0.5f);
			   glMultMatrixf((float *) &farPMat);
			   f = nearZ * tan(E3M_TORAD(modifyPointScroll3->getValue()/2.0f));
			   glBegin(GL_TRIANGLE_STRIP);
			      glVertex3f(-f, f, nearZ);
			      glVertex3f(f, f, nearZ);
			      glVertex3f(-f, -f, nearZ);
			      glVertex3f(f, -f, nearZ);
			   glEnd();
			   glPopMatrix();
			   glDepthMask(true);			   
			MVISIO::end3D();			
			
			MVISIO::begin3D(cameraOrto);			   
			   root->pass();			   
			MVISIO::end3D();	
			
			farZ = -modifyPointScroll2->getValue();
		   nearZ = -modifyPointScroll1->getValue();			
			f = farZ * tan(E3M_TORAD(modifyPointScroll3->getValue()/2.0f));
			//MVISIO::clear(false, true, false);
			glDisable(GL_LIGHTING);
			MVISIO::begin3D(cameraOrto);			   
			   glPushMatrix();
			   //glDepthMask(false);			
			   glMultMatrixf((float *) &root->getFinalMatrix());  
			   glDisable(GL_CULL_FACE);
			   glEnable(GL_BLEND);
			   glColor4f(1.0f, 1.0f, 0.0f, 0.5f);
			   glMultMatrixf((float *) &farPMat);
			    
			   glBegin(GL_TRIANGLE_STRIP);
			      glVertex3f(-f, f, farZ);
			      glVertex3f(f, f, farZ);
			      glVertex3f(-f, -f, farZ);
			      glVertex3f(f, -f, farZ);
			   glEnd();
			   glPopMatrix();
			   
			   glPushMatrix();			
			   glMultMatrixf((float *) &root->getFinalMatrix());  
			   glDisable(GL_CULL_FACE);
			   glEnable(GL_BLEND);			   
			   glColor4f(1.0f, 1.0f, 0.0f, 0.5f);
			   glMultMatrixf((float *) &farPMat);
			   f = nearZ * tan(E3M_TORAD(modifyPointScroll3->getValue()/2.0f));
			   glBegin(GL_TRIANGLE_STRIP);
			      glVertex3f(-f, f, nearZ);
			      glVertex3f(f, f, nearZ);
			      glVertex3f(-f, -f, nearZ);
			      glVertex3f(f, -f, nearZ);
			   glEnd();
			   glPopMatrix();
			  // glDepthMask(true);			   
			MVISIO::end3D();					
			MVVECTOR4 _fA = MVCAMERA::getCameraMatrix()->rotate(fA);					
			
			// Normal viewport:
			MVOPENGL::resize(0, 0, MODULE_SIZEX, MODULE_SIZEY);
			
			// Foreground:						
			MVISIO::begin2D();		
			   PASS_QUITBUTTON;
			   // Title:
			   font->putString(10, 13, MODULE_NAME);	  				  
			   glDisable(GL_TEXTURE_2D);
			   glLineWidth(LINE_WIDTH);
			   
			   glBegin(GL_LINES);			      
			      glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
			      glVertex2i(MODULE_SIZEX/2.0f, MODULE_SIZEY/18.2);				      
			      glVertex2i(MODULE_SIZEX/2.0f, MODULE_SIZEY-MODULE_SIZEY/8.1);
			   glEnd();
			   interfaceWin->pass();
			MVISIO::end2D();
			
			
			////////////////
			// Check events:			
			SDL_Event event;			
			//SDL_PumpEvents();						
			while (SDL_PollEvent(&event)) 
			{
					// Quit by click on X:
					if (event.type == SDL_QUIT) 
							done = true;	
               
               // Quit by pressing ESC:
					if (event.type == SDL_KEYDOWN) 
					{
							// Quit by ESC:
							if (event.key.keysym.sym == SDLK_ESCAPE) 
							       done = true;															
			      }															
			}         
			
			// Update keyboard (like a joypad):
			Uint8 *array;
			int arraySize;
			array = SDL_GetKeyState(&arraySize);
			if (array[SDLK_PAGEUP])    
			{
               depth += 2.0f * MVISIO::getFPSfactor();  
               if (depth > 40.0f)
			            depth = 40.0f;			             
         }							       								                     
         else
               if (array[SDLK_PAGEDOWN])
               {
                     depth -= 2.0f * MVISIO::getFPSfactor();
                       if (depth < -40.0f)
			                  depth = -40.0f;                                          
               }							      
			
			float deltaX = mouseX - oldMouseX;
         float deltaY = mouseY - oldMouseY;		
         oldMouseX = mouseX; oldMouseY = mouseY;	
         oldMouseBL = mouseBL;	
			
		   // Update mouse:
			int buttonState = SDL_GetMouseState(&mouseX, &mouseY);		
         if (buttonState&SDL_BUTTON(1))
					mouseBL = true;					
			else 
					mouseBL = false;
         if (buttonState&SDL_BUTTON(2) || buttonState&SDL_BUTTON(3))
					mouseBR = true;					
			else 
					mouseBR = false;					
			
			// Check MVisio events:
			MVID id=0;
			occurred = MVELEMENT::manageGUIEvents(0, mouseX, mouseY, mouseBL, mouseBR, &id);  
			if (PRESSED_QUITBUTTON(id))
			      done = true;

         // Rotate scene:
         if (!occurred && mouseX<MODULE_SIZEX/2)
         {
               if (mouseBL && !mouseBR)
               {
                     // Rotate model:
			            rotationX += deltaY/4.0f;
			            rotationY += deltaX/4.0f;
			            if (rotationX > 68.0f) rotationX = 68.0f;
			            if (rotationX < -68.0f) rotationX = -68.0f;			         
			      }  
			      if (mouseBL && mouseBR)
			      {
			            depth += deltaY/7.0f;
			            if (depth < -40.0f)
			                  depth = -40.0f;
                     if (depth > 40.0f)
			                  depth = 40.0f;			                  			               
			      }                   
         }			   
         
         if (id==resetButton->getId())
         {
               modifyPointScroll1->setValue(1.0f);
               modifyPointScroll2->setValue(50.0f);
               modifyPointScroll3->setValue(45.0f);
               rotationX = 0.0f;
               rotationY = 90.0f;
               depth = 10.0f;               
         }   
			
			// Display:		
			Sleep(10); // Don't kill mthreading ;)
			MVISIO::swap();					
	}      
   
   // Done:
   MVISIO::free();
   return 0;
}